@@ -14,13 +14,13 @@ Sub::Exporter - a sophisticated exporter for custom-built routines
 
 =head1 VERSION
 
-version 0.966
+version 0.974
 
-  $Id: /my/cs/projects/export/trunk/lib/Sub/Exporter.pm 22590 2006-06-17T19:54:13.510080Z rjbs  $
+  $Id: /my/cs/projects/Sub-Exporter/trunk/lib/Sub/Exporter.pm 31455 2007-04-22T13:38:51.095739Z rjbs  $
 
 =cut
 
-our $VERSION = '0.966';
+our $VERSION = '0.974';
 
 =head1 SYNOPSIS
 
@@ -176,7 +176,8 @@ The following keys are valid in C<%config>:
   exports - a list of routines to provide for exporting; each routine may be
             followed by generator
   groups  - a list of groups to provide for exporting; each must be followed by
-            a list of exports, possibly with arguments for each export
+            either (a) a list of exports, possibly with arguments for each
+            export, or (b) a generator
   collectors - a list of names into which values are collected for use in
                routine generation; each name may be followed by a validator
 
@@ -190,13 +191,13 @@ The list is processed in such a way that the following are equivalent:
   { exports =>
     { foo => undef, bar => undef, baz => undef, quux => \&quux_generator } }
 
-Generators are coderefs that return coderefs.  They are called with four
+Generators are code that return coderefs.  They are called with four
 parameters:
 
   $class - the class whose exporter has been called (the exporting class)
   $name  - the name of the export for which the routine is being build
  \%arg   - the arguments passed for this export
- \%coll  - the collections for this import
+ \%col   - the collections for this import
 
 Given the configuration in the L</SYNOPSIS>, the following C<use> statement:
 
@@ -216,6 +217,11 @@ would result in the following call to C<&build_reformatter>:
 The returned coderef (C<$code>) would then be installed as C<make_narrow> in the
 calling package.
 
+Instead of providing a coderef in the configuration, a reference to a method
+name may be provided.  This method will then be called on the invocant of the
+C<import> method.  (In this case, we do not pass the C<$class> parameter, as it
+would be redundant.)
+
 =head2 Group Configuration
 
 The C<groups> list can be passed in the same forms as C<exports>.  Groups must
@@ -372,7 +378,7 @@ thrown.
 sub _group_name {
   my ($name) = @_;
 
-  return if (index '-:', (substr $name, 0, 1)) == -1;
+  return if (index q{-:}, (substr $name, 0, 1)) == -1;
   return substr $name, 1;
 }
 
@@ -400,7 +406,11 @@ sub _expand_groups {
       my $prefix = (delete $merge{-prefix}) || '';
       my $suffix = (delete $merge{-suffix}) || '';
 
-      if (Params::Util::_CODELIKE($groups[$i][1])) {
+      if (
+        Params::Util::_CODELIKE($groups[$i][1]) ## no critic Private
+        or
+        Params::Util::_SCALAR0($groups[$i][1]) ## no critic Private
+      ) {
         # this entry was build by a group generator
         $groups[$i][0] = $prefix . $groups[$i][0] . $suffix;
       } else {
@@ -443,10 +453,25 @@ sub _expand_group {
 
   my $exports = $config->{groups}{$group_name};
 
-  if (Params::Util::_CODELIKE($exports)) {
-    my $group = $exports->($class, $group_name, $group_arg, $collection);
+  if (
+    Params::Util::_CODELIKE($exports) ## no critic Private
+    or
+    Params::Util::_SCALAR0($exports) ## no critic Private
+  ) {
+    # I'm not very happy with this code for hiding -prefix and -suffix, but
+    # it's needed, and I'm not sure, offhand, how to make it better.
+    # -- rjbs, 2006-12-05
+    my $group_arg = $group_arg ? { %$group_arg } : {};
+    delete $group_arg->{-prefix};
+    delete $group_arg->{-suffix};
+
+    my $group = Params::Util::_CODELIKE($exports)
+              ? $exports->($class, $group_name, $group_arg, $collection)
+              : $class->$$exports($group_name, $group_arg, $collection);
+
     Carp::croak qq(group generator "$group_name" did not return a hashref)
       if ref $group ne 'HASH';
+
     my $stuff = [ map { [ $_ => $group->{$_} ] } keys %$group ];
     return @{
       _expand_groups($class, $config, $stuff, $collection, $seen, $merge)
@@ -490,8 +515,12 @@ sub _collect_collections {
         into        => $into,
       };
 
-      Carp::croak "collection $name failed validation"
-        unless $hook->($value, $arg);
+      my $error_msg = "collection $name failed validation";
+      if (Params::Util::_SCALAR0($hook)) {
+        Carp::croak $error_msg unless $class->$$hook($value, $arg);
+      } else {
+        Carp::croak $error_msg unless $hook->($value, $arg);
+      }
     }
   }
 
@@ -585,16 +614,27 @@ sub _rewrite_build_config {
   Carp::croak q(into and into_level may not both be supplied to exporter)
     if exists $config->{into} and exists $config->{into_level};
 
-  $config->{$_} = Data::OptList::mkopt_hash($config->{$_}, $_, 'CODE')
-    for qw(exports collectors);
+  for (qw(exports collectors)) {
+    $config->{$_} = Data::OptList::mkopt_hash(
+      $config->{$_},
+      $_,
+      [ 'CODE', 'SCALAR' ],
+    );
+  }
 
   if (my @names = _key_intersection(@$config{qw(exports collectors)})) {
     Carp::croak "names (@names) used in both collections and exports";
   }
 
-  $config->{groups}
-    = Data::OptList::mkopt_hash(
-      $config->{groups}, 'groups', [ 'HASH', 'CODE', 'ARRAY' ]
+  $config->{groups} = Data::OptList::mkopt_hash(
+      $config->{groups},
+      'groups',
+      [
+        'HASH',   # standard opt list
+        'ARRAY',  # standard opt list
+        'CODE',   # group generator
+        'SCALAR', # name of group generation method
+      ]
     );
 
   # by default, export nothing
@@ -631,9 +671,11 @@ sub build_exporter {
     # this builds a AOA, where the inner arrays are [ name => value_ref ]
     my $import_args = Data::OptList::mkopt([ @_ ]);
 
+    # is this right?  defaults first or collectors first? -- rjbs, 2006-06-24
+    $import_args = [ [ -default => undef ] ] unless @$import_args;
+
     my $collection = _collect_collections($config, $import_args, $class, $into);
 
-    $import_args = [ [ -default => 1 ] ] unless @$import_args;
     my $to_import = _expand_groups($class, $config, $import_args, $collection);
 
     # now, finally $import_arg is really the "to do" list
@@ -650,7 +692,7 @@ sub _do_import {
 
   my ($generator, $as);
 
-  if ($arg and Params::Util::_CODELIKE($arg)) {
+  if ($arg and Params::Util::_CODELIKE($arg)) { ## no critic
     # This is the case when a group generator has inserted name/code pairs.
     $generator = sub { $arg };
     $as = $name;
@@ -713,12 +755,21 @@ sub default_exporter {
 sub _generate {
   my ($class, $generator, $name, $arg, $collection) = @_;
 
-  # I considered making the T case, below, "$class->$generator(" but it seems
-  # that overloading precedence would turn an overloaded-as-code generator
-  # object into a string before code. -- rjbs, 2006-06-11
-  my $code = $generator
-           ? $generator->($class, $name, $arg, $collection)
-           : $class->can($name); 
+  if (not defined $generator) {
+    my $code = $class->can($name)
+      or Carp::croak "can't locate exported subroutine $name via $class";
+    return $code;
+  }
+
+  # I considered making this "$class->$generator(" but it seems that
+  # overloading precedence would turn an overloaded-as-code generator object
+  # into a string before code. -- rjbs, 2006-06-11
+  return $generator->($class, $name, $arg, $collection)
+    if Params::Util::_CODELIKE($generator);
+
+  # This "must" be a scalar reference, to a generator method name.
+  # -- rjbs, 2006-12-05
+  return $class->$$generator($name, $arg, $collection);
 }
 
 sub _install {
@@ -798,8 +849,8 @@ complex uses, Sub::Exporter makes hard things possible, which would not be
 possible with Exporter. 
 
 When using a module that uses Sub::Exporter, users familiar with Exporter will
-probably see difference in the basics.  These two lines do about the same thing
-in whether the exporting module uses Exporter or Sub::Exporter.
+probably see no difference in the basics.  These two lines do about the same
+thing in whether the exporting module uses Exporter or Sub::Exporter.
 
   use Some::Module qw(foo bar baz);
   use Some::Module qw(foo :bar baz);
